from abc import ABC, abstractmethod
from .resource import Item, Collection, Null


class Transformer(ABC):
    includes = []

    @abstractmethod
    def transform(self, source):
        raise NotImplementedError

    def item(self, source, transformer):
        return Item(source, transformer)

    def collection(self, source, transformer):
        return Collection(source, transformer)

    def null(self):
        return Null()



import json


class Resource(object):
    def __init__(self, data, transformer):
        self.data = data
        self.transformer = transformer


class Item(Resource):
    def apply_transformer(self):
        if not self.data:
            return Result(None)

        result = self.transformer.transform(self.data)

        for include in self.transformer.includes:
            method_name = 'include_{}'.format(include)
            if not hasattr(self.transformer, method_name):
                raise ResourceException('Transformer must contain method: {}'.format(method_name))
            method_instance = getattr(self.transformer, method_name)
            include_result = method_instance(self.data).apply_transformer()
            result[include] = include_result.get_data()

        return Result(result)


class Collection(Resource):
    def apply_transformer(self):
        if not self.data:
            return Result([])

        result = []

        for value in self.data:
            item = Item(value, self.transformer)
            result.append(item.apply_transformer().get_data())

        return Result(result)


class Null(Resource):
    def __init__(self):
        super(Null, self).__init__(None, None)

    def apply_transformer(self):
        return Result(None)


class Result(object):
    def __init__(self, data):
        self.data = data

    def to_json(self):
        return json.dumps(self.data)

    def get_data(self):
        return self.data


class ResourceException(Exception):
    pass



import pytest
from collections import namedtuple
from fractal_output import Manager, Transformer, Item, Collection, ResourceException


class TestTransformer(object):
    def test_transform_item(self):
        User = namedtuple('User', 'first_name')

        class UserTransformer(Transformer):
            def transform(self, user):
                return {
                    'firstName': user.first_name
                }

        user = User('abigail')
        transformer = UserTransformer()
        item = Item(user, transformer)

        manager = Manager()
        result = manager.create_data(item).to_json()

        assert '{"firstName": "abigail"}' == result

    def test_transform_collection(self):
        User = namedtuple('User', 'first_name')

        class UserTransformer(Transformer):
            def transform(self, user):
                return {
                    'firstName': user.first_name
                }

        users = [User('abigail'), User('mary')]
        transformer = UserTransformer()
        item = Collection(users, transformer)

        manager = Manager()
        result = manager.create_data(item).to_json()

        assert '[{"firstName": "abigail"}, {"firstName": "mary"}]' == result

    def test_transform_item_empty(self):
        class UserTransformer(Transformer):
            def transform(self, user):
                return {
                    'firstName': user.first_name
                }

        transformer = UserTransformer()
        item = Item(None, transformer)

        manager = Manager()
        result = manager.create_data(item).get_data()

        assert not result

    def test_transform_collection_empty(self):
        class UserTransformer(Transformer):
            def transform(self, user):
                return {
                    'firstName': user.first_name
                }

        transformer = UserTransformer()
        item = Collection(None, transformer)

        manager = Manager()
        result = manager.create_data(item).to_json()

        assert '[]' == result

    def test_transform_includes(self):
        User = namedtuple('User', 'first_name telephones address')
        Telephone = namedtuple('Telephone', 'number')
        Address = namedtuple('Address', 'street')

        class TelephoneTransformer(Transformer):
            def transform(self, telephone):
                return {
                    'number': telephone.number
                }

        class AddressTransformer(Transformer):
            def transform(self, address):
                return {
                    'street': address.street
                }

        class UserTransformer(Transformer):
            includes = [
                'telephones',
                'address',
                'subscription'
            ]

            def transform(self, user):
                return {
                    'firstName': user.first_name
                }

            def include_address(self, user):
                return self.item(user.address, AddressTransformer())

            def include_telephones(self, user):
                return self.collection(user.telephones, TelephoneTransformer())

            def include_subscription(self, user):
                return self.null()

        user = User(
            first_name='abigail',
            telephones=[Telephone('1')],
            address=Address('lala')
        )

        transformer = UserTransformer()
        item = Item(user, transformer)

        manager = Manager()
        result = manager.create_data(item).get_data()

        assert 'abigail' == result['firstName']
        assert '1' == result['telephones'][0]['number']
        assert 'lala' == result['address']['street']
        assert not result['subscription']

    def test_invalid_include(self):
        User = namedtuple('User', 'first_name address')
        Address = namedtuple('Address', 'street')

        class AddressTransformer(Transformer):
            def transform(self, address):
                return {
                    'street': address.street
                }

        class UserTransformer(Transformer):
            includes = [
                'lala'
            ]

            def transform(self, user):
                return {
                    'firstName': user.first_name
                }

            def include_address(self, user):
                return self.item(user.address, AddressTransformer())

        user = User(
            first_name='abigail',
            address=Address('lala')
        )

        transformer = UserTransformer()
        item = Item(user, transformer)
        manager = Manager()

        with pytest.raises(ResourceException):
            manager.create_data(item)



from setuptools import setup
import fractal_output

long_description = open('README.rst', 'r').read()

setup(
    name='fractal_output',
    version=fractal_output.__version__,
    packages=['fractal_output'],
    setup_requires=['wheel'],
    install_requires=[],
    entry_points={
        "console_scripts": [
            "fractal_output = fractal_output.__main__:__main__"
        ],
    },
    description="The task manager for python",
    long_description=long_description,
    url='https://github.com/jefersondaniel/fractal-output',
    author='Jeferson Daniel',
    author_email='jeferson.daniel412@gmail.com',
    license='MIT',
    classifiers=[
        'Operating System :: OS Independent',
        'Development Status :: 5 - Production/Stable',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ]
)
